from django.utils.translation import gettext_lazy as _
import copy

def make_name(name: str):
    name = name.replace(" ", "_")
    name = name.replace("-", "_")
    name = name.lower()
    return name

class CookieConsentPanelMeta(type):
    def __new__(cls, name, bases, attrs):

        if attrs.get("abstract", False):
            return super().__new__(cls, name, bases, attrs)
        
        if "name" not in attrs:
            attrs["name"] = make_name(name)
        
        return super().__new__(cls, name, bases, attrs)

class CookieConsentPanel(metaclass=CookieConsentPanelMeta):
    abstract    = True
    name        = None
    title       = None
    description = None
    # will be used on the cookie policy page.
    long_description = None
    # Is the field required?
    required    = False
    # Will always be true if the field is required.
    initial     = False

    def __init__(self, request):
        self.request = request

    def __eq__(self, other: "CookieConsentPanel"):
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, CookieConsentPanel):
            return self.name == other.name
        else:
            raise TypeError(f"Cannot compare {self.__class__.__name__} with {other.__class__.__name__}")

cookie_registry: dict[str, CookieConsentPanel] = {}

class RegisteredPanel:
    def __init__(self, panel: CookieConsentPanel, order: int):
        self.panel = panel
        self.order = order

    def __eq__(self, other: "RegisteredPanel"):
        if isinstance(other, RegisteredPanel):
            return self.panel.name == other.panel.name
        elif isinstance(other, CookieConsentPanel) or (type(other) == type and issubclass(other, CookieConsentPanel)):
            return self.panel.name == other.name
        else:
            raise TypeError(f"Cannot compare {self.__class__.__name__} with {other.__class__.__name__}")
    
    def __lt__(self, other: "RegisteredPanel"):
        if isinstance(other, RegisteredPanel):
            return self.order < other.order
        return self.order < other
    
    def __gt__(self, other: "RegisteredPanel"):
        if isinstance(other, RegisteredPanel):
            return self.order > other.order
        return self.order > other

def register_cookie_panel(panel: CookieConsentPanel=None, order: int = 0):
    if panel is None:
        return lambda panel: register_cookie_panel(panel, order)
    
    if not issubclass(panel, CookieConsentPanel):
        raise TypeError(f"Expected CookieConsentPanel, got {panel.__class__.__name__}")
    
    if panel.name in cookie_registry:
        # remove the panel to change the order
        del cookie_registry[panel.name]
    
    cookie_registry[panel.name] = RegisteredPanel(panel, order)

def unregister_cookie_panel(panel: CookieConsentPanel):
    name = panel
    if hasattr(panel, "name"):
        name = panel.name

    if name in cookie_registry:
        del cookie_registry[name]

def get_cookie_panels(order=True):
    panels: list[RegisteredPanel] = copy.deepcopy(list(cookie_registry.values()))

    if order:
        # Sort by registered panels.
        panels.sort()

    for panel in panels:
        # Yield the actual panel.
        yield panel.panel

def init_cookie_panels(request) -> list[CookieConsentPanel]:
    cookie_panels = get_cookie_panels()
    cookie_panels = [panel(request) for panel in cookie_panels]
    return cookie_panels




from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from wagtail import hooks

class ModalExtra:
    def is_shown(self, request):
        return True

    def render(self, request, context):
        raise NotImplementedError()
    
class AnchorTagExtra(ModalExtra):
    get_url: callable  = None
    get_text: callable = None
    classes: str       = ""

    def __init__(self, get_url=None, get_text=None, classes=""):
        self.get_url  = get_url
        self.get_text = get_text
        self.classes  = classes

    def render(self, request, context):
        if callable(self.get_url):
            url = self.get_url(self.request)
        else:
            url = self.get_url

        if callable(self.get_text):
            text = self.get_text(self.request)
        else:
            text = self.get_text

        return mark_safe(f'<a href="{url}" class="{self.classes}">{text}</a>')
    
class RowModalExtra(ModalExtra):
    columns: list[ModalExtra] = []

    def render(self, request, context):
        return mark_safe(
            '<div class="row">' +
            "".join(
                f'<div class="col">{column.render(request, context)}</div>'
                for column in self.columns
            ) +
            '</div>'
        )
        


# We will check if wagtail is installed first; their hook system is simple; 
# and keeps everything centralized.
try:
    try:
        from wagtail.core import hooks
    except ImportError:
        from wagtail import hooks

    def register_cookie_modal_extra(extra: ModalExtra):
        hooks.register("register_cookies_modal_extras")(extra)

    def get_cookie_modal_extras(request):
        return hooks.get_hooks("register_cookies_modal_extras")

except ImportError:
    # If wagtail is not installed, we will try to get the extras from the
    # django settings. The string defined in settings should point to a
    # callable that returns a list of ModalExtras, or a list of ModalExtras.
    from django.conf import settings
    from django.utils.module_loading import import_string

    try:
        _ = settings.COOKIE_CONSENT_EXTRAS

        def register_cookie_modal_extra(extra: ModalExtra):
            raise NotImplementedError("Cannot register cookie modal extras when using django settings to define them")

        def get_cookie_modal_extras(request):
            extras = getattr(settings, "COOKIE_CONSENT_EXTRAS", [])
            imported = import_string(extras)
            if callable(imported):
                return imported(request)
            return imported

    except AttributeError:
        modal_extras = []

        def register_cookie_modal_extra(extra: ModalExtra):
            modal_extras.append(extra)

        def get_cookie_modal_extras(request):
            return modal_extras





from django.template import Library, loader, Node, TemplateSyntaxError
from django.urls import reverse

register = Library()

from ..forms import (
    make_consent_form,
)

from ..options import (
    COOKIE_CONSENT_LOGO,
    COOKIE_CONSENT_LOGO_URL,
    COOKIE_CONSENT_TITLE,
    COOKIE_CONSENT_DESCRIPTION,
    COOKIE_CONSENT_BUTTON_ACCEPT,
    COOKIE_CONSENT_BUTTON_ACCEPT_ALL,
    COOKIE_CONSENT_BUTTON_SETTINGS,
    COOKIE_CONSENT_BUTTON_DECLINE,
)

from ..cookies import (
    ACCEPT_ALL_COOKIE,
)

from ..extras import (
    get_cookie_modal_extras,
)

class CookieConsentNode(Node):
    def __init__(self, nodelist, args):
        self.nodelist = nodelist
        self.args = args
    
    def render(self, context):
        request = context["request"]
        args = []
        for arg in self.args:
            args.append(arg.resolve(context))

        if request.cookie_consent.allows(*args):
            return self.nodelist.render(context)
        
        return ""

@register.tag(name="consents_to")
def do_consents_to(parser, token):

    end_tag = "endconsents_to"
    args = []

    bits = token.split_contents()
    tag_name = bits.pop(0)

    if len(bits) < 1:
        raise TemplateSyntaxError(f"{tag_name} requires at least one argument")
    
    while len(bits) > 0:
        arg = bits.pop(0)
        if arg == "as":
            break

        compiled = parser.compile_filter(arg)
        args.append(compiled)

    nodelist = parser.parse((end_tag,))
    parser.delete_first_token()

    return CookieConsentNode(nodelist, args)

@register.simple_tag(takes_context=True)
def render_with_context(context, item):
    request = context["request"]
    if hasattr(item, "render"):
        return item.render(request, context)
    return item

@register.simple_tag(takes_context=True)
def cookie_consent(context):
    request = context["request"]

    action_url = reverse("cookie_consent:submit")
    policy_url = reverse("cookie_consent:policy")

    if request.path.lower().strip("/") == policy_url.lower().strip("/"):
        return ""

    if not request.cookie_consent.submitted:

        cookie_ctx = {
            "request": request,
            "parent_context": context,

            "cookie_modal_logo":        COOKIE_CONSENT_LOGO,
            "cookie_modal_logo_url":    COOKIE_CONSENT_LOGO_URL,
            "cookie_modal_title":       COOKIE_CONSENT_TITLE,
            "cookie_modal_description": COOKIE_CONSENT_DESCRIPTION,
            "cookie_modal_form_action": action_url,
            "cookie_modal_policy_url":  policy_url,
            "cookie_modal_form":        make_consent_form(request),
            "cookie_modal_extras":      get_cookie_modal_extras(request),

            "cookie_modal_accept_all_value": ACCEPT_ALL_COOKIE,

            "cookie_modal_button_accept":     COOKIE_CONSENT_BUTTON_ACCEPT,
            "cookie_modal_button_accept_all": COOKIE_CONSENT_BUTTON_ACCEPT_ALL,
            "cookie_modal_button_settings":   COOKIE_CONSENT_BUTTON_SETTINGS,
            "cookie_modal_button_decline":    COOKIE_CONSENT_BUTTON_DECLINE,
        }

        template = loader.get_template("cookie_consent/cookie_modal.html")
        return template.render(cookie_ctx, request)

    return ""




from django.shortcuts import render
from django.http import (
    JsonResponse,
    HttpResponseRedirect,
)
from .panels import (
    init_cookie_panels,
)
from .forms import (
    make_consent_form_class,
    set_cookie_consent_submitted,
)
from .cookies import (
    delete_cookie_consent,
)
from .options import (
    COOKIE_CONSENT_POLICY_PAGE_EXTENDS_FROM,
)

cookie_policy_panels = []

class CookiePolicyPanel:
    title: str = ""
    description: str = ""

    def __eq__(self, other):
        if isinstance(other, CookiePolicyPanel):
            return self.title == other.title
        elif isinstance(other, str):
            return self.title == other
        else:
            raise TypeError(f"Cannot compare {self.__class__.__name__} with {other.__class__.__name__}")

def register_cookie_policy_panel(panel: CookiePolicyPanel):
    cookie_policy_panels.append(panel)

def unregister_cookie_policy_panel(panel: CookiePolicyPanel):
    cookie_policy_panels.remove(panel)

def revoke_cookie_consent(request):
    set_cookie_consent_submitted(request, False)

    panels = init_cookie_panels(request)
    for panel in panels:
        delete_cookie_consent(request, panel.name)

    if "X-Requested-With" in request.headers:
        return JsonResponse(
            {
                "success": True, 
            }, 
            status=200
        )
    
    return HttpResponseRedirect(request.META.get("HTTP_REFERER", "/"))

def submit_cookie_consent_form(request):

    if not request.method == "POST":
        return JsonResponse({"error": "Invalid request method"}, status=400)
    
    panels = init_cookie_panels(request)
    form_cls = make_consent_form_class(panels)
    form = form_cls(request, request.POST)

    if not form.is_valid():
        return JsonResponse({"error": "Invalid form"}, status=400)
    
    form.save()

    if "X-Requested-With" in request.headers:
        return JsonResponse(
            {
                "success": True, 
            }, 
            status=200
        )
    
    return HttpResponseRedirect(request.META.get("HTTP_REFERER", "/"))

def policy(request):
    panels = init_cookie_panels(request)

    panels = [panel for panel in panels if panel.long_description]

    context = {
        "panels": panels,
        "extends_from": COOKIE_CONSENT_POLICY_PAGE_EXTENDS_FROM,
        "extra_panels": cookie_policy_panels,
    }

    return render(request, "cookie_consent/cookie_policy.html", context)



from django.urls import path
from .views import (
    submit_cookie_consent_form,
    revoke_cookie_consent,
    policy,
)

app_name = "cookie_consent"

urlpatterns = [
    path("submit/", submit_cookie_consent_form, name="submit"),
    path("revoke/", revoke_cookie_consent, name="revoke"),
    path("policy/", policy, name="policy"),
]


from typing import Any
from django import forms
from django.forms.widgets import Widget
from django.utils.translation import gettext_lazy as _

from .cookies import (
    ACCEPT_ALL_COOKIE,
    store_cookie_consent,
    set_cookie_consent_submitted,
)

from .panels import (
    init_cookie_panels,
    CookieConsentPanel,
)

class CookieConsentForm(forms.Form):
    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.all_accepted = False

    def full_clean(self) -> None:
        self.data = self.data.copy()
        for field, value in self.fields.items():
            if value.required:
                self.data[field] = "on"
    
        if ACCEPT_ALL_COOKIE in self.data:
            accept_all = self.data[ACCEPT_ALL_COOKIE]
            if accept_all.lower() == ACCEPT_ALL_COOKIE.lower():
                self.all_accepted = True
                
        super().full_clean()

    def save(self):
        print(self.cleaned_data)

        if self.all_accepted:
            for field in self.fields:
                store_cookie_consent(self.request, field, True)
        else:
            for field in self.fields:
                store_cookie_consent(self.request, field, self.cleaned_data[field])
        
        set_cookie_consent_submitted(self.request, True)
        self.request.session.modified = True
        return self.cleaned_data

class BooleanField(forms.BooleanField):
    def widget_attrs(self, widget: Widget) -> Any:
        if self.required:
            return {"disabled": ""}
        return {}
    
    def clean(self, value):
        if self.required:
            return True
        return super().clean(value)

def make_consent_form_class(cookie_panels: list[CookieConsentPanel]):
    fields = {
        panel.name: BooleanField(
            label=panel.title,
            required=panel.required,
            initial=panel.required or panel.initial,
            help_text=panel.description,
        )
        for panel in cookie_panels
    }

    return type("CookiePanelForm", (CookieConsentForm,), fields)

def make_consent_form(request, data=None):
    cookie_panels = init_cookie_panels(request)
    form_class = make_consent_form_class(cookie_panels)
    return form_class(request, data=data)



ACCEPT_ALL_COOKIE = "cookie_consent_accept_all"
HAS_SUBMITTED_COOKIE = "cookie_consent_submitted"

def make_cookie_name(name: str):
    name = name.replace(" ", "_")
    name = name.replace("-", "_")
    name = name.lower()
    return f"cookie_consent_{name}"

def _is_true(value: str):
    if isinstance(value, bool):
        return value
    return value.lower() in ["true", "1", "yes", "on"]

def has_consent(request, *cookie_name):
    if not hasattr(request, "session"):
        return False

    consents = True

    for name in cookie_name:
        name = make_cookie_name(name)
        value = request.session.get(name, None)
        if value is None:
            return False
            
        consents = consents and _is_true(value)
        if not consents:
            return False
        
    return consents

def cookie_consent_submitted(request):
    if not hasattr(request, "session"):
        return False
    has_accepted_consent = request.session.get(HAS_SUBMITTED_COOKIE, None)
    if has_accepted_consent is None:
        return False
    
    return _is_true(has_accepted_consent)

def set_cookie_consent_submitted(request, value: bool):
    if not hasattr(request, "session"):
        return False
    request.session[HAS_SUBMITTED_COOKIE] = value

def store_cookie_consent(request, cookie_name, cookie_value):
    request.session[make_cookie_name(cookie_name)] = cookie_value

def delete_cookie_consent(request, cookie_name):
    if not hasattr(request, "session"):
        return False
    if cookie_name in request.session:
        del request.session[cookie_name]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

from .panels import (
    CookieConsentPanel,
    register_cookie_panel,
)

from .extras import (
    AnchorTagExtra,
    register_cookie_modal_extra,
)

class FunctionalCookieConsentPanel(CookieConsentPanel):
    name        = "functional"
    title       = _("Functional cookies")
    description = _("These cookies are required for the website to function.")
    required    = True
    long_description = _(
        """Functional cookies are required for the website to function.
They are usually only set in response to actions made by you which amount to a request for services,
such as setting your privacy preferences, logging in or filling in forms.
You can set your browser to block or alert you about these cookies,
but some parts of the site will not then work.
These cookies do not store any personally identifiable information.
""")


class AnalyticsCookieConsentPanel(CookieConsentPanel):
    name        = "analytics"
    title       = _("Analytics cookies")
    description = _("This site uses analytics cookies to improve your experience.")
    required    = False
    long_description = _(
        """Analytics cookies allow us to count visits and traffic sources,
so we can measure and improve the performance of our site.
They help us know which pages are the most and least popular
and see how visitors move around the site.
All information these cookies collect is aggregated and therefore anonymous.
If you do not allow these cookies,
we will not know when you have visited our site.
""")


class MarketingCookieConsentPanel(CookieConsentPanel):
    name        = "marketing"
    title       = _("Marketing cookies")
    description = _("This site uses marketing cookies to improve your experience.")
    required    = False
    long_description = _(
        """Marketing cookies are used to track visitors across websites.
The intention is to display ads that are relevant and engaging for the individual user
and thereby more valuable for publishers and third party advertisers.
If you do not allow these cookies,
you will experience less targeted advertising.
""")

class CookieConsentConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cookie_consent'

    def ready(self) -> None:
        from django.urls import reverse_lazy

        register_cookie_modal_extra(AnchorTagExtra(
            get_text=_("Cookie Policy"),
            get_url=reverse_lazy("cookie_consent:policy"),
            classes="link",
        ))

        register_cookie_panel(FunctionalCookieConsentPanel, 0)
        register_cookie_panel(AnalyticsCookieConsentPanel, 1)
        register_cookie_panel(MarketingCookieConsentPanel, 2)





from .cookies import (
    cookie_consent_submitted,
    has_consent,
)

class CookieConsent:

    def __init__(self, request):
        self.request = request

    def allows(self, *cookie_name):
        if not cookie_name:
            return False
        return has_consent(self.request, *cookie_name)
    
    @property
    def submitted(self):
        return cookie_consent_submitted(self.request)


class CookieConsentMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):

        request.cookie_consent = CookieConsent(request)

        response = self.get_response(request)

        return response


from django.conf import settings
from django.utils.translation import gettext_lazy as _
import os

COOKIE_CONSENT_LOGO = getattr(settings, "COOKIE_CONSENT_LOGO", "cookie_consent/img/ga-logo.svg")
COOKIE_CONSENT_LOGO_URL = getattr(settings, "COOKIE_CONSENT_LOGO_URL", "https://goodadvice.it")
COOKIE_CONSENT_TITLE = getattr(settings, "COOKIE_CONSENT_TITLE", _("Cookies"))
COOKIE_CONSENT_DESCRIPTION = getattr(settings, "COOKIE_CONSENT_DESCRIPTION", _("This site uses cookies to improve your experience."))
COOKIE_CONSENT_BUTTON_ACCEPT = getattr(settings, "COOKIE_CONSENT_BUTTON_ACCEPT", _("Accept"))
COOKIE_CONSENT_BUTTON_ACCEPT_ALL = getattr(settings, "COOKIE_CONSENT_BUTTON_ACCEPT_ALL", _("Accept All"))
COOKIE_CONSENT_BUTTON_DECLINE = getattr(settings, "COOKIE_CONSENT_BUTTON_DECLINE", _("Decline"))
COOKIE_CONSENT_BUTTON_SETTINGS = getattr(settings, "COOKIE_CONSENT_BUTTON_SETTINGS", _("Settings"))
COOKIE_CONSENT_POLICY_PAGE_EXTENDS_FROM = getattr(settings, "COOKIE_CONSENT_POLICY_PAGE_EXTENDS_FROM", "base.html")

